use super::sql_json::SqlJson;
use super::DbConn;
use sprinklers_core::{
    model::{
        Program, ProgramId, ProgramItem, ProgramSequence, ProgramUpdateData, Programs, SectionId,
    },
    schedule::Schedule,
};

use eyre::Result;
use rusqlite::{params, Row, ToSql, Transaction, NO_PARAMS};

type SqlProgramSequence = SqlJson<ProgramSequence>;
type SqlSchedule = SqlJson<Schedule>;

fn from_sql<'a>(row: &Row<'a>) -> rusqlite::Result<Program> {
    Ok(Program {
        id: row.get(0)?,
        name: row.get(1)?,
        enabled: row.get(2)?,
        schedule: row.get::<_, SqlSchedule>(3)?.into_inner(),
        sequence: row.get::<_, SqlProgramSequence>(4)?.into_inner(),
    })
}

struct SqlProgramUpdate<'a> {
    id: ProgramId,
    name: Option<&'a String>,
    enabled: Option<bool>,
    schedule: Option<SqlJson<&'a Schedule>>,
}

impl<'a> IntoIterator for &'a SqlProgramUpdate<'a> {
    type Item = &'a dyn ToSql;
    type IntoIter = std::vec::IntoIter<Self::Item>;

    fn into_iter(self) -> Self::IntoIter {
        vec![
            &self.id as &dyn ToSql,
            &self.name,
            &self.enabled,
            &self.schedule,
        ]
        .into_iter()
    }
}

fn update_as_sql(id: ProgramId, program: &ProgramUpdateData) -> SqlProgramUpdate {
    SqlProgramUpdate {
        id,
        name: program.name.as_ref(),
        enabled: program.enabled,
        schedule: program.schedule.as_ref().map(SqlJson),
    }
}

struct SqlProgramItem {
    program_id: ProgramId,
    seq_num: isize,
    section_id: SectionId,
    duration: f64,
}

impl<'a> IntoIterator for &'a SqlProgramItem {
    type Item = &'a dyn ToSql;
    type IntoIter = std::vec::IntoIter<Self::Item>;

    fn into_iter(self) -> Self::IntoIter {
        vec![
            &self.program_id as &dyn ToSql,
            &self.seq_num,
            &self.section_id,
            &self.duration,
        ]
        .into_iter()
    }
}

fn item_as_sql(
    program_item: &ProgramItem,
    program_id: ProgramId,
    seq_num: usize,
) -> SqlProgramItem {
    SqlProgramItem {
        program_id,
        seq_num: (seq_num + 1) as isize,
        section_id: program_item.section_id,
        duration: program_item.duration.as_secs_f64(),
    }
}

fn sequence_as_sql<'a>(
    program_id: ProgramId,
    sequence: &'a [ProgramItem],
) -> impl Iterator<Item = SqlProgramItem> + 'a {
    sequence
        .iter()
        .enumerate()
        .map(move |(seq_num, item)| item_as_sql(item, program_id, seq_num))
}

pub fn query_programs(conn: &DbConn) -> Result<Programs> {
    let query_sql = "\
SELECT p.id, p.name, p.enabled, p.schedule, ps.sequence
FROM programs AS p
         INNER JOIN program_sequences AS ps ON ps.program_id = p.id;";
    let mut statement = conn.prepare_cached(query_sql)?;
    let rows = statement.query_map(NO_PARAMS, from_sql)?;
    let mut programs = Programs::new();
    for row in rows {
        let program = row?;
        programs.insert(program.id, program.into());
    }
    Ok(programs)
}

pub fn query_program_by_id(conn: &DbConn, id: ProgramId) -> Result<Program> {
    let query_sql = "\
SELECT p.id, p.name, p.enabled, p.schedule, ps.sequence
FROM programs AS p
         INNER JOIN program_sequences AS ps ON ps.program_id = p.id
WHERE p.id = ?1;";
    let mut statement = conn.prepare_cached(query_sql)?;
    Ok(statement.query_row(params![id], from_sql)?)
}

pub fn update_program(
    trans: &mut Transaction,
    id: ProgramId,
    prog: &ProgramUpdateData,
) -> Result<()> {
    let save = trans.savepoint()?;
    let conn = &*save;
    let update_sql = "\
UPDATE programs
    SET name = ifnull(?2, name),
        enabled = ifnull(?3, enabled),
        schedule = ifnull(?4, schedule)
    WHERE id = ?1;";
    conn.prepare_cached(update_sql)?
        .execute(&update_as_sql(id, prog))?;
    if let Some(sequence) = &prog.sequence {
        let clear_seq_sql = "\
DELETE
FROM program_sequence_items
WHERE program_id = ?1;";
        conn.prepare_cached(clear_seq_sql)?.execute(params![id])?;
        let insert_seq_sql = "\
INSERT INTO program_sequence_items (program_id, seq_num, section_id, duration)
VALUES (?1, ?2, ?3, ?4);";
        let mut insert_seq = conn.prepare_cached(insert_seq_sql)?;
        for params in sequence_as_sql(id, sequence) {
            insert_seq.execute(&params)?;
        }
        drop(insert_seq);
    }
    save.commit()?;
    Ok(())
}
